import re
import logging


# create a logger for this module
logger = logging.getLogger(__name__)

# Some no commands are not forgiving, e.g., "no license boot level
# network-advantage addon dna-advantage" is rejected, but "no license boot
# level" is acceptable. These cases are listed in SHORT_NO_COMMANDS.
SHORT_NO_COMMANDS = [
    'no license boot level ',
    'no transport input ',
    'no transport output ',
]

# Some commands are orderless, e.g., "show running-config" output could be:
# aaa authentication login admin-con group tacacs+ local
# aaa authentication login admin-vty group tacacs+ local
# Or in other times it is displayed in a different order:
# aaa authentication login admin-vty group tacacs+ local
# aaa authentication login admin-con group tacacs+ local
ORDERLESS_COMMANDS = [
    r'^ *aaa authentication login ',
]

# Some commands can be overwritten without a no command. For example, changing
# from:
# username admin privilege 15 password 0 admin
# to:
# username admin privilege 15 password 7 15130F010D24
# There is no need to send a no command before sending the second line.
OVERWRITABLE_COMMANDS = [
    r'^ *username \S+ privilege [0-9]+ password ',
    r'^ *password ',
    r'^ *description ',
]

# Some commands look like a parent-child relation but actually they are
# siblings. One example is two lines of client config below:
# aaa server radius proxy
#  client 10.0.0.0 255.0.0.0
#   !
#   client 11.0.0.0 255.0.0.0
#   !
SIBLING_CAMMANDS = [
    r'^ *client ',
]

# As for the client command above, its children does not have indentation:
# aaa server radius proxy
#  client 10.0.0.0 255.0.0.0
#   timer disconnect acct-stop 23
#   !
#   client 11.0.0.0 255.0.0.0
#   accounting port 34
#   timer disconnect acct-stop 22
#   !
#   client 12.0.0.0 255.0.0.0
#   accounting port 56
#   !
# Logically, "accounting port 34" and "timer disconnect acct-stop 22" are
# children of "client 11.0.0.0 255.0.0.0", but there is no indentation in the
# "show running-config" output. The sub-section is indicated by the expression
# mark.
MISSING_INDENT_COMMANDS = [
    r'^ *client ',
]


class ListDiff(object):
    '''ListDiff

    Abstraction of diff between two lists. Each item in the list is a tuple.
    If all tuples are same, two lists are considered to be same. This class is
    intended to be used by class RunningConfigDiff.

    Attributes
    ----------
    diff : `list`
        A list whose items are tuples. The first item in the tuple is
        the config line and the second item is its next level config. If a
        config line does not have next level config, the second item value is
        None. Otherwise, the value is a list of tuples. The third item in the
        tuple has three possible values: '+' means the config is added; '-'
        means it is deleted; '' means it remiains unchnaged but for some reason
        it is needed (e.g., as a position reference).
    '''

    def __init__(self, list1, list2):
        self.list1 = list1
        self.list2 = list2

    @property
    def diff(self):
        return self.compare(self.list1, self.list2)

    @staticmethod
    def compare(list1, list2):
        diff_1 = []
        key_list_1 = [i[0] for i in list1]
        key_list_2 = [i[0] for i in list2]
        common_key_list = [k for k in key_list_1 if k in key_list_2]
        if common_key_list:
            common_1 = list(filter(lambda x: x in common_key_list, key_list_1))
            common_2 = list(filter(lambda x: x in common_key_list, key_list_2))
            common_key_list = []
            list1_index = 0
            for k2 in common_2:
                if k2 in common_1[list1_index:]:
                    common_key_list.append(k2)
                    list1_index = common_1.index(k2)

        previous_index_1 = previous_index_2 = 0
        for key in common_key_list:
            current_index_1 = key_list_1.index(key)
            current_index_2 = key_list_2.index(key)

            # Stuff in list1 before a common key but not in list2
            for k, v, i in list1[previous_index_1:current_index_1]:
                diff_1.append((k, v, '-'))

            # Stuff in list2 before a common key but not in list1
            for k, v, i in list2[previous_index_2:current_index_2]:
                diff_1.append((k, v, '+'))

            # Stuff of the common key itself
            if list1[current_index_1][1] == list2[current_index_2][1]:
                if list1[current_index_1][1] is None:
                    diff_1.append((key, None, '?'))
                else:
                    diff_1.append((key, [
                        (k, v, '')
                        for k, v, i in list1[current_index_1][1]], '?'))
            else:
                if (
                    list1[current_index_1][1] is not None and
                    list2[current_index_2][1] is not None
                ):
                    diff_1.append((key, ListDiff.compare(
                        list1[current_index_1][1],
                        list2[current_index_2][1],
                    ), ''))
                elif list1[current_index_1][1] is not None:
                    diff_1.append((key, [
                        (k, v, '-')
                        for k, v, i in list1[current_index_1][1]], ''))
                elif list2[current_index_2][1] is not None:
                    diff_1.append((key, [
                        (k, v, '+')
                        for k, v, i in list2[current_index_2][1]], ''))

            previous_index_1 = current_index_1 + 1
            previous_index_2 = current_index_2 + 1

        # Stuff after all common keys
        for k, v, i in list1[previous_index_1:]:
            diff_1.append((k, v, '-'))
        for k, v, i in list2[previous_index_2:]:
            diff_1.append((k, v, '+'))

        # Cleanup
        diff_2 = []
        keys_before = set()
        for index, item in enumerate(diff_1):
            key, value, info = item
            if info == '?':
                keys_after = set([k for k, v, i in diff_1[index+1:]])
                if keys_before & keys_after:
                    diff_2.append((key, value, ''))
            else:
                diff_2.append((key, value, info))
            keys_before.add(key)

        return diff_2


class RunningConfigDiff(object):
    '''RunningConfigDiff

    Abstraction of diff between two Cisco running-configs. It supports str()
    which returns a string showing the differences between running1 and
    running2.

    Attributes
    ----------
    running1 : `str`
        First Cisco running-config.

    running2 : `str`
        Second Cisco running-config.

    diff : `list`
        A list from class ListDiff attribute diff.

    cli : `str`
        CLIs that transits from running1 to running2.

    cli_reverse : `str`
        CLIs that transits from running2 to running1.
    '''

    def __init__(self, running1, running2):
        '''
        __init__ instantiates a RunningConfigDiff instance.
        '''

        self.running1 = running1
        self.running2 = running2
        self._diff_list = None

    def __bool__(self):
        return bool(self.diff)

    def __str__(self):
        return self.list2config(self.diff, diff_type='')

    def __eq__(self, other):
        if str(self) == str(other):
            return True
        else:
            return False

    @property
    def diff(self):
        list1 = self.running2list(self.running1)
        list2 = self.running2list(self.running2)
        self.handle_sibling_cammands(list1)
        self.handle_sibling_cammands(list2)
        self._diff_list = ListDiff(list1, list2).diff
        return self._diff_list if self._diff_list else None

    @property
    def cli(self):
        return self.get_cli(reverse=False)

    @property
    def cli_reverse(self):
        return self.get_cli(reverse=True)

    def get_cli(self, reverse=False):
        if self._diff_list is None:
            self.diff
        positive_str, negative_str = self.list2cli(
            self._diff_list, reverse=reverse)
        if positive_str:
            if negative_str:
                return negative_str + '\n!\n' + positive_str
            else:
                return positive_str
        else:
            return negative_str

    def running2list(self, str_in):
        str_in = str_in.replace('exit-address-family', ' exit-address-family')
        return self.config2list(self.handle_orderless(str_in))

    def config2list(self, str_in):
        list_ret = []
        last_line = ''
        last_section = ''
        last_indentation = 0
        missing_indent = False
        for line in str_in.splitlines():
            if len(line.strip()) > 22 and \
               line[:22] == 'Building configuration':
                continue
            if len(line.strip()) > 21 and \
               line[:21] == 'Current configuration':
                continue
            if len(line.strip()) == 0:
                continue
            if missing_indent and line.rstrip() == '!':
                if last_line:
                    if last_indentation > 0:
                        list_ret.append((
                            last_line, self.config2list(last_section), ''))
                    else:
                        list_ret.append((last_line, None, ''))
                last_line = ''
                last_section = ''
                last_indentation = 0
                missing_indent = False
                continue
            if re.search('^ *!', line) or re.search('^ *%', line):
                continue
            if line[0] == ' ':
                current_indentation = len(re.search('^ *', line).group(0))
                if last_indentation == 0:
                    last_indentation = current_indentation

                # There might be special cases. For example, the following
                # running-config:
                # ip dhcp class CLASS1
                #    relay agent information
                #  relay-information hex 01040101030402020102
                # should be considered as:
                # ip dhcp class CLASS1
                #    relay agent information
                #     relay-information hex 01040101030402020102
                if current_indentation < last_indentation:
                    last_section += ' ' * (last_indentation + 1) + \
                        line[current_indentation:] + '\n'
                else:
                    last_section += line[last_indentation:] + '\n'
            else:
                if any(
                    [re.search(regx, line) for regx in MISSING_INDENT_COMMANDS]
                ):
                    missing_indent = True
                elif missing_indent:
                    current_indentation = 1
                    if last_indentation == 0:
                        last_indentation = current_indentation
                    last_section += line + '\n'
                    continue
                if last_line:
                    if last_indentation > 0:
                        list_ret.append((
                            last_line, self.config2list(last_section), ''))
                    else:
                        list_ret.append((last_line, None, ''))
                last_line = line
                last_section = ''
                last_indentation = 0
        if last_line:
            if last_indentation > 0:
                list_ret.append((
                    last_line, self.config2list(last_section), ''))
            else:
                list_ret.append((last_line, None, ''))
        return list_ret

    def list2config(self, list_in, diff_type=None):
        str_ret = ''
        if list_in is None:
            return str_ret
        for k, v, i in list_in:
            if k == '':
                continue
            if diff_type == '':
                local_diff_type = i
            else:
                local_diff_type = diff_type
            if diff_type is None:
                str_ret += '  ' + k + '\n'
            else:
                prefix = local_diff_type if local_diff_type != '' else ' '
                str_ret += prefix + ' ' + k + '\n'
            if v is not None:
                str_ret += self.indent(
                    self.list2config(v, diff_type=local_diff_type),
                )
        return str_ret

    def list2cli(self, list_in, reverse=False):
        if list_in is None:
            return ''
        plus = '-' if reverse else '+'
        minus = '+' if reverse else '-'
        positive_list = []
        negative_list = []
        positive_keys = [k for k, v, i in list_in if i == plus and v is None]
        for k, v, i in list_in:
            if k == '':
                continue
            if v is None:
                if i == plus:
                    cmd, is_no_cmd = self.handle_commands(k)
                    if is_no_cmd:
                        negative_list.append(cmd)
                    else:
                        positive_list.append(cmd)
                elif i == minus:
                    # In a case that a CLI is updated, no command is not
                    # needed:
                    # - service timestamps debug datetime msec
                    # + service timestamps debug datetime msec localtime
                    #   show-timezone
                    key_len = len(k)
                    matching_positive_keys = [
                        key for key in positive_keys if key[:key_len] == k]
                    if not matching_positive_keys:
                        cmd, is_no_cmd = self.handle_commands('no ' + k)
                        if is_no_cmd:
                            negative_list.append(cmd)
                        else:
                            positive_list.append(cmd)
            else:
                if i == '':
                    positive_str, negative_str = self.list2cli(
                        v, reverse=reverse)
                    if positive_str:
                        positive_list.append(
                            k + '\n' + self.indent(positive_str).rstrip())
                    if negative_str:
                        negative_list.append(
                            k + '\n' + self.indent(negative_str).rstrip())
                if i == plus:
                    positive_str = self.list2config(v, diff_type=None).rstrip()
                    if positive_str:
                        positive_list.append(k + '\n' + positive_str)
                    else:
                        positive_list.append(k)
                elif i == minus:
                    cmd, is_no_cmd = self.handle_commands('no ' + k)
                    if is_no_cmd:
                        negative_list.append(cmd)
                    else:
                        positive_list.append(cmd)

        # Handle overwritable commands
        for regx in OVERWRITABLE_COMMANDS:
            for cmd in positive_list:
                m = re.search(regx, cmd)
                if m:
                    exact_cmd = 'no ' + m.group(0).lstrip()
                    len_exact_cmd = len(exact_cmd)
                    for idx, line in enumerate(negative_list):
                        cmd_line = line.lstrip()
                        if (
                            len(cmd_line) >= len_exact_cmd and
                            cmd_line[:len_exact_cmd] == exact_cmd
                        ):
                            break
                    else:
                        continue
                    del negative_list[idx]

        # Handle duplicate commands
        # Some commands their positive and negative lines are both appeared in
        # the "show running-config" output, e.g., "platform punt-keepalive
        # disable-kernel-core" and "no platform punt-keepalive
        # disable-kernel-core" are both visible. This treatment is required to
        # avoid two lines of the same CLI.
        for cmd_list in [positive_list, negative_list]:
            self.remove_duplicate_cammands(cmd_list)

        return '\n'.join(positive_list), '\n'.join(reversed(negative_list))

    @staticmethod
    def handle_commands(cmd):
        if cmd[:6] == 'no no ':
            return cmd[6:], False
        for short_cmd in SHORT_NO_COMMANDS:
            if cmd[:len(short_cmd)] == short_cmd:
                return short_cmd, True
        if cmd[:3] == 'no ':
            return cmd, True
        else:
            return cmd, False

    @staticmethod
    def handle_orderless(running):
        lines = running.splitlines()
        for regx in ORDERLESS_COMMANDS:
            indexes = []
            for idx, line in enumerate(lines):
                if re.search(regx, line):
                    indexes.append(idx)
            if len(indexes) < 2:
                continue
            matches_dict = {hash(lines[i]): lines[i] for i in indexes}
            matches_list = [
                matches_dict[i] for i in sorted(list(matches_dict.keys()))]
            for i in reversed(indexes):
                del lines[i]
            lines = lines[:indexes[0]] + matches_list + lines[indexes[0]:]
        return '\n'.join(lines)

    @staticmethod
    def remove_duplicate_cammands(config_list):
        indexes = []
        commands = set()
        for idx, line in enumerate(config_list):
            if line in commands:
                indexes.append(idx)
            else:
                commands.add(line)
        if indexes:
            for idx in reversed(indexes):
                del config_list[idx]

    @staticmethod
    def handle_sibling_cammands(config_list):
        length = len(config_list)
        lines_inserted = 0
        for i in range(length):
            idx = i + lines_inserted
            tup = config_list[idx]
            for regx in SIBLING_CAMMANDS:
                if re.search(regx, tup[0]) and tup[1] is not None:
                    siblings = []
                    indexes = []
                    for idx_c, tup_c in enumerate(tup[1]):
                        if re.search(regx, tup_c[0]):
                            indexes.append(idx_c)
                    for idx_c in reversed(indexes):
                        siblings.append(tup[1][idx_c])
                        del tup[1][idx_c]
                    if not tup[1]:
                        tup = config_list[idx] = (tup[0], None, tup[2])
                    j = 1
                    for sibling in reversed(siblings):
                        config_list.insert(i+j, sibling)
                        j += 1
                    break
            if tup[1] is not None:
                RunningConfigDiff.handle_sibling_cammands(tup[1])

    def indent(self, str_in):
        str_ret = ''
        for line in str_in.splitlines():
            if line:
                if line[0] in '-+':
                    diff_type = line[0]
                    line = line[1:]
                    str_ret += diff_type + '  ' + line + '\n'
                else:
                    str_ret += '  ' + line + '\n'
        return str_ret
